// Copyright 2002 Adam Megacz, see the COPYING file for licensing [GPL]
package org.xwt;

import org.xwt.util.*;
import java.util.*;
import org.mozilla.javascript.*;

/**
 *  This class encapsulates a single trap placed on a given node. The
 *  traps for a given property name on a given box are maintained as a
 *  linked list stack, with the most recently placed trap at the head
 *  of the list.
 */
public class Trap {

    // Static Data //////////////////////////////////////////////////////////////

    /** a vector of weak references to all instances of Trap; used for retheming */
    private static Hashtable allTraps = new Hashtable(1000);

    /** List of properties that cannot be trapped */
    private static final Hash PROHIBITED = new Hash(120, 3);

    static {
        String[] p = new String[] {
            "sizetoimage", "shrink", "hshrink", "vshrink", "x", "y", "width", "height",
            "flex", "align", "invisible", "absolute", "globalx", "globaly",
            "minwidth", "minheight", "height", "width", "maxwidth", "maxheight", 
            "numchildren", "hpad", "vpad", "doublebuffered", "cursor",
            "mousex", "mousey", "xwt", "static", "mouseinside", "root", "thisbox", "indexof", "svg"
        };
        for(int i=0; i<p.length; i++) PROHIBITED.put(p[i], Boolean.TRUE);
    };

    private static Object dummy = new Object();


    // Instance Members ////////////////////////////////////////////////////////

    /** the box on which this trap was placed */
    private Box trapee = null;

    /** If this trap was actually placed on a root proxy, this is a reference to the proxy */
    private Scriptable rp = null;
    
    /** the function for this trap */
    Function f = null;

    /** the name of the property that this trap was placed on */
    private String name = null;

    /** the next trap down the trap stack */
    private Trap next = null;

    /** true if this is a read-trap */
    boolean isreadtrap = false;

    /** The nodeName of the Template whose script placed the trap. See Template.nodeName for more detail. */
    private String placerNodeName = null;

    /** our weak reference in the allTraps Hash */
    private Weak myWeak = Platform.getWeak(this);


    // Static Methods //////////////////////////////////////////////////////////////////////////

    /**
     *  adds a trap.
     *  @param trapee the box to place the trap on
     *  @param name the name of the property to trap on
     *  @param f the function to place as a trap
     *  @param isreadtrap true iff this is a read (double-underscore) trap
     *  @param rp if this trap is being placed via a rootproxy, this is that proxy object.
     */
    static void addTrap(Box trapee, String name, Function f, boolean isreadtrap, Scriptable rp) {

        if (PROHIBITED.get(name) != null || name.startsWith("xwt_")) {
            System.out.println("Error: you cannot place traps on special property \"" + name + "\"");
            return;
        }

        // find out what script is currently running
        String placerNodeName = JSObject.getCurrentFunctionSourceName();

        // check if this script has already placed a trap on this property
        if (trapee.traps == null) trapee.traps = new Hash(10, 3);
        for(Trap t = (Trap)trapee.traps.get(name); t != null; t = t.next) if (t.placerNodeName.equals(placerNodeName)) return;
        
        // actually place the trap
        Trap t = new Trap();
        t.next = (Trap)trapee.traps.get(name);
        trapee.traps.put(name, t);
        t.trapee = trapee;
        t.f = f;
        t.placerNodeName = placerNodeName;
        t.rp = rp;
        t.name = name;
        t.isreadtrap = isreadtrap;

        // make sure that the surface knows when we place traps on these two properties (slightly inelegant)
        if (trapee.surface != null && (name.equals("KeyPressed") || name.equals("KeyReleased"))) {
            trapee.surface.keywatchers.removeElement(trapee);
            trapee.surface.keywatchers.addElement(trapee);
        }
    }

    /** returns the function placed as a trap on property <code>name</code> of box <code>b</code> by the currently-running script */
    public static Trap getTrap(Box b, String name) {
        if (b.traps == null) return null;

        String currentFunctionNodeName = JSObject.getCurrentFunctionSourceName();
        for(Trap cur = (Trap)b.traps.get(name); cur != null; cur = cur.next)
            if (cur.placerNodeName.equals(currentFunctionNodeName))
                return cur;

        return null;
    }

    /** Called by Rhino's arguments.cascade. Note: cx will be null if this was invoked from perform() rather than from a script. */
    public static final Function cascadeFunction = new CascadeFunction();
    private static class CascadeFunction extends JSObject implements Function {
        CascadeFunction() { setSeal(true); }
        public Scriptable construct(Context cx, Scriptable scope, java.lang.Object[] args) { return null; }
        public Object call(Context cx, Scriptable thisObj, Scriptable ctorObj, Object[] args) {
            Trap currentTrap = TrapContext.get().currentTrap;
            if (currentTrap == null || (cx != null && Context.getCurrentContext().currentFunction != currentTrap.f)) {
                if (Log.on) Log.log(this, "attempt to cascade() by a function that was not invoked as a trap at " +
                                    cx.interpreterSourceFile + ":" + cx.interpreterLine);
                return null;
            }
            if (args.length != 0) TrapContext.get().putCascadeHappened = true;
            Trap t = currentTrap.next;
            // if we've hit the end of the trap stack, just do a put(,,,true)
            if (t == null) {
                if (args.length == 0) return currentTrap.trapee.get(currentTrap.name, currentTrap.trapee, true);
                currentTrap.trapee.put(currentTrap.name, currentTrap.trapee, args[0], true);
                return null;
            }
            return t.perform(args);
        }
    };

    /** called by Rhino's arguments.trapee hack */
    public static Object currentTrapee() {
        Trap current = TrapContext.get().currentTrap;
        if (current == null) return null;
        else if (current.rp != null) return current.rp;
        else return current.trapee;
    }

    /** called by Rhino's arguments.trapname hack */
    public static String currentTrapname() {
        Trap current = TrapContext.get().currentTrap;
        if (current == null) return null;
        else return current.name;
    }

    /** removes all traps whose function's ultimate parent scope is <tt>b</tt>. Used for retheming */
    public static void removeAllTrapsByBox(Box b) {
        Enumeration e = allTraps.keys();
        while(e.hasMoreElements()) {
            Weak w = (Weak)e.nextElement();
            Trap t = (Trap)w.get();
            if (t == null) {
                allTraps.remove(w);
                continue;
            }
            for(Scriptable cur = t.f; cur != null; cur = cur.getParentScope()) {
                if (cur == b) {
                    t.delete();
                    break;
                }
            }
        }
    }


    // Instance Methods //////////////////////////////////////////////////////////////////////////

    private Trap() { allTraps.put(myWeak, dummy); }

    /** perform this trap -- arg.length == 0 if this is a get; otherwise it contains a single element to be put */
    public Object perform(Object[] arg) {
        TrapContext tc = TrapContext.get();

        // save both thread-locals on the stack and update their values
        Trap save_currentTrap = tc.currentTrap;
        tc.currentTrap = this;

        boolean save_putCascadeHappened = tc.putCascadeHappened; 
        tc.putCascadeHappened = false;

        // invoke the trap function
        try {
            if (!isreadtrap && arg.length == 0) return cascadeFunction.call(null, null, null, arg);

            if (f == null) {
                if (Log.verbose) Log.log(this, "debug: reclaimed a dangling trap on property " + name);
                Object ret = cascadeFunction.call(null, null, null, arg);
                delete();
                return ret;
            }
            
            Object ret = f.call(Context.enter(), f.getParentScope(), f.getParentScope(), arg);
            
            // autocascade if required
            if (arg.length > 0 && !isreadtrap && !tc.putCascadeHappened) cascadeFunction.call(null, null, null, arg);
            
            return ret;

        } catch (EcmaError e) {
            if (Log.on) Log.log(this, "WARNING: uncaught interpreter exception: " + e.getMessage());
            if (Log.on) Log.log(this, "         thrown from within trap '" + name + "' at " + e.getSourceName() + ":" + e.getLineNumber());
        } catch (JavaScriptException e) {
            if (Log.on) Log.log(this, "WARNING: uncaught ecmascript exception: " + e.getMessage());
            if (Log.on) Log.log(this, "         thrown from within trap '" + name + "' at " + e.sourceFile + ":" + e.line);
        } finally {
            // restore the thread-locals
            tc.putCascadeHappened = save_putCascadeHappened;
            tc.currentTrap = save_currentTrap;
            tc.trapDepth--;
        }
        return null;
    }

    /** removes this trap */
    public void delete() {
        for(Trap last = null, cur = (Trap)trapee.traps.get(name); cur != null; last = cur, cur = cur.next) {
            if (cur != this) continue;
            else if (last != null) last.next = cur.next;
            else if (cur.next == null) trapee.traps.remove(name);
            else trapee.traps.put(name, cur.next);
        }
        if (trapee.surface != null && !trapee.is_trapped("KeyPressed") && !trapee.is_trapped("KeyReleased"))
            trapee.surface.keywatchers.removeElement(trapee);
        allTraps.remove(myWeak);
    }

    /** per-thread storage for Traps */
    private static class TrapContext {

        private static Hash trapContextByThread = new Hash();
        private TrapContext() { }

        private boolean putCascadeHappened = false;
        private Trap currentTrap = null;
        private int trapDepth = 0;

        /** returns the TrapContext for the current thread */
        static TrapContext get() {
            TrapContext ret = (TrapContext)trapContextByThread.get(Thread.currentThread());
            if (ret == null) {
                ret = new TrapContext();
                trapContextByThread.put(Thread.currentThread(), ret);
            }
            return ret;
        }

    }

}

